//---------------------------------------------------------------------------------------
// Контроллер шагового двигателя Nema17 c драйвером A4988 на Arduino Pro Mini 5V/16MHz
// Файл stepper_17.09.15.ino
// ZagAl - на основе информации предоставленной OldBean.
//---------------------------------------------------------------------------------------
#include <AccelStepper.h> Библиотека для шагового двигателя с "external driver" (A4988)
#include <Wire.h>
#define SLAVE_ADDRESS       0x07        // Адрес шины i2c для связи с внешним миром. Slave.
//---------------------------------------------------------------------------------------
#define COMMON_CATHODE // Если индикатор с общим анодом - необходимо закомментировать эту строку
//---------------------------------------------------------------------------------------
#ifdef COMMON_CATHODE // Для индикаторов с общим катодом
#define L_HIGH 1
#define L_LOW  0
#else                 // Для индикаторов с общим анодом
#define L_HIGH 0
#define L_LOW  1
#endif
//---------------------------------------------------------------------------------------
// Константы - состояние кнопок
#define NO_ACT              0
#define PLUS_CLICK          0b00000001  // Короткое нажатие кнопки "+"
#define MINUS_CLICK         0b00000010  // Короткое нажатие кнопки "-"
#define PLUS_PRESS          0b00000100  // Удержание кнопки "+" в нажатом состоянии
#define MINUS_PRESS         0b00001000  // Удержание кнопки "-" в нажатом состоянии

//---------------------------------------------------------------------------------------
static uint8_t mButPin = 0, // Пин, к которому подключена кнопка "-"
               pButPin = 1, // Пин, к которому подключена кнопка "+"
               outPin = 2,  // Пин выходного сигнала (для управления спящим режимом. Подключить к pin - ENABLE на драйвере A4988.)
               msPin = 3;  // Пин для соединения с MS1 - MS3 на драйвере A4988 для дробления шага.           
//---------------------------------------------------------------------------------------
// Установки, связанные с индикатором, рассчитаны на трехразрядные семисегментные индикаторы
// с общими катодами (или анодами). Выводы анодов (или катодов для индикаторов с общими анодами) 
// подключены непосредственно к соответствующим пинам МК. Выводы общих катодов (общих анодов для
// индикаторов с общими анодами) подключены через резисторы 470 Ом.
//---------------------------------------------------------------------------------------
// Пины, к которым подключены сегменты индикатора - аноды (или катоды для индикаторов с общими анодами)
static uint8_t segPins[7] = {12, 16, 5, 7, 8, 13, 4};
// Пины, к которым подключены общие катоды (или аноды, для индикаторов с общими анодами)
static uint8_t digPins[3] = {15, 14, 11};
static uint8_t digit[11] = {0b00111111, // Маски сегментов цифр 0, 1, 2 и т.д.
                            0b00000110,
                            0b01011011,
                            0b01001111,
                            0b01100110,
                            0b01101101,
                            0b01111101,
                            0b00000111,
                            0b01111111,
                            0b01101111,
                            0b00000000}; // "Цифра" 10 это пробел
static int d[3]; // Сотни, десятки и единицы показываемые на разрядах 7-сегментного индикатора
static uint8_t mask = 0x01; // Маска для анализа цифр
static int seg = 0; // Номер сегмента, обрабатываемого (подствечиваемого, если нужно) в данный момент
static volatile int cntr = 0; // Счетчик для тактирования смены подствечиваемых сегментов
//---------------------------------------------------------------------------------------
static int dtBounce = 10; // Задержка (мс) для устранения дребезга контактов кнопок
static int dtPress = 300; // Задержка, после которой считается, что кнопка удерживается
static uint8_t bos = NO_ACT; // Предыдущее состояние кнопок
static unsigned long tpr; // Момент нажатия кнопки
static uint8_t stp = 0; // Начальное положение шагового двигателя
static uint8_t stpChangedFlag = 1;
static uint8_t dSTP = 5;  // Приращение устанавливаемой кнопками мощности при удержании кнопок нажатыми

static uint8_t pFlag = 0; // Наличие сигнала (импульса) на выходе модулятора
static unsigned long pStart; // Момент начала импульса на выходе модулятора
static unsigned long T = 1000; // Задержка включения мотора при выходе из спящего режима - в мс.
static unsigned long pWidth = 30000; // Длительность ожидания перед уходом в спящий режим - в мс.
//---------------------------------------------------------------------------------------
// Define a stepper and the pins it will use
AccelStepper stepper(1,10,9);  // Использует пин 10 для step и 9 для dir, 1 - режим "external driver" (A4988)
static uint8_t step_division = 4;  // Деление шага в соответствии с установленной перемычкой на A4988.
//---------------------------------------------------------------------------------------
void setup() {
  Wire.begin(SLAVE_ADDRESS); // Инициализация библиотеки для шины i2c
  Wire.onReceive(receiveData);
  Wire.onRequest(sendData);
  stepper.setMaxSpeed(30.0); // Cкорость шагового двигателя.
  stepper.setAcceleration(100.0); // Ускорение шагового двигателя.

  pinMode(mButPin, INPUT); // Инициализация пинов, к которым подключены кнопки
  digitalWrite(mButPin, HIGH);
  pinMode(pButPin, INPUT);
  digitalWrite(pButPin, HIGH);

  pinMode(outPin, OUTPUT); // Иницализация пина выходного сигнала (соединен с ENABLE на A4988)
  digitalWrite(outPin, LOW);

  pinMode(msPin, OUTPUT); // Иницализация пина выходного сигнала (соединен с MS1-MS3 на A4988)
  digitalWrite(msPin, HIGH);

  for(int i = 0; i < 7; i++) { // Инициализация пинов, к которым подключены сегменты индикатора
    pinMode(segPins[i], OUTPUT);
    digitalWrite(segPins[i], L_LOW);
  }
  for(int i = 0; i < 3; i++) { // Инициализация пинов, к которым подключены общие катоды (аноды) индикатора
    pinMode(digPins[i], OUTPUT);
    digitalWrite(digPins[i], L_HIGH);
  }
  unsigned long st = millis() + 1000;
  d[0] = 8; d[1] = 8; d[2] = 8; // Для демонстрации работоспособности индикатора покажем все сегменты разом
  while(millis() < st) {
      ShowSegments();
  }
  d[0] = 10; d[1] = 10; d[2] = 10; // Очистим индикатор
  ShowSegments();

  // Инициализация таймера 2 для тактирования системы
  TCCR2A = (1<<WGM21);            // Режим CTC (сброс по совпадению)
  TCCR2B = (1<<CS21) | (1<<CS20); // Предделитель 32
  OCR2A = 98;                     // При этом значении совпадения происходят с частотой около 5 кГц.
  TIMSK2 = (1<<OCIE2A);           // Разрешить прерывание по совпадению
}
//---------------------------------------------------------------------------------------
ISR (TIMER2_COMPA_vect) { // Обработчик прерывания от таймера TIMER2

// Анализируем состояние кнопок
  uint8_t ks = GetKey();
  switch(ks) {
    case PLUS_CLICK:  if(stp >= 200) stp = 200; // За один оборот шаговый двигатель в режиме полного шага делает 200 шагов.
                      else stp++;
                      stpChangedFlag = 1;
                      break;
    case MINUS_CLICK: if(stp == 0) stp = 0;
                      else stp--;
                      stpChangedFlag = 1;
                      break;
    case PLUS_PRESS:  if(stp >= 200 - dSTP) stp = 200;
                      else stp += dSTP;
                      stpChangedFlag = 1;
                      break;
    case MINUS_PRESS: if(stp <= dSTP) stp = 0;
                      else stp -= dSTP;
                      stpChangedFlag = 1;
                      break;
  }
  if(stpChangedFlag) { // Скорректируем показания индикатора, если stp был изменен
    UpdateDigits(stp);
    pStart=0; // Обнуляем период активации A4988  
  }
  if(++cntr >= 10) { // Каждые 2 мс переходим к подсветке следующей группы сегментов
    cntr = 0; ShowSegments();
  }
  unsigned long now = millis();
  if(pFlag==0 && stpChangedFlag==1) { // Если был спящий режим и произошли изменения от клавиш или i2c
    digitalWrite(outPin, LOW);        // то выводим из спящего режима A4988 - шаговый двигатель может работать.
    pStart = now; pFlag = 1;          // засекаем время активации A4988 и ставим флаг активации.
    stpChangedFlag=0;
  }
  if(pFlag==1 && now - pStart > pWidth) { // Если установлен флаг активации и/но период активации истек,
    digitalWrite(outPin, HIGH);           // вводим A4988 в спящий режим.
    pFlag = 0;                            // Флаг активации обнуляем.
  }
}
//---------------------------------------------------------------------------------------
void loop() {
    if(stepper.distanceToGo()==0){ // Если двигатель достиг нужной позиции, то
    stepper.runToNewPosition((200-stp)*step_division); // начать движение к новой позиции рассчитанной Малинкой или установленой кнопками с учетом деления шага двигателя.
  }                                    //(200-stp) меняет направление вращения двигателя.
}
//---------------------------------------------------------------------------------------
uint8_t GetKey(void) {
  uint8_t ret, bcs = ((!digitalRead(mButPin) & 0x01) << 1) | (!digitalRead(pButPin) & 0x01); // Текущее состояние кнопок
  if(!bos && bcs) { // Произошло нажатие кнопки
    tpr = millis(); bos = bcs; // Фиксируем время и состояние кнопок
    return NO_ACT;
  }
  if(bos && !bcs) { // Произошло либо отжатие или дребезг кнопки
    if((millis() - tpr) < dtBounce) return NO_ACT; // Дребезг, однако...
    ret = bos; bos = bcs; return ret; // Отжатие...
  }
  if(bos && bcs) { // Возможно, удерживание кнопки
    if((millis() - tpr) < dtPress) return NO_ACT; // Но пока еще не ясно...
    // И все-таки это удержание кнопки в нажатом состоянии
    ret = bos << 2; tpr = millis(); return ret;
  }
  return NO_ACT;
}
//---------------------------------------------------------------------------------------
void UpdateDigits(int val) { // Подготовка данных (цифр) для визуализации
  d[2] = val/100;
  d[1] = (val - d[2]*100)/10;
  d[0] = val - d[2]*100 - d[1]*10;
  if(d[2] == 0) { // Не отображаем ведущие нули ежели что... ("Цифра" 10 - это "пустая" цифра, все ее сегменты потушены)
    d[2] = 10;
    if(d[1] == 0) d[1] = 10;
  }
}
//---------------------------------------------------------------------------------------
void ShowSegments() { // Подсвечивание очередной группы сегментов
  // Комментарии написаны для индикаторов с общими катодами. В скобках - для индикаторов с общими анодами
  digitalWrite(segPins[seg], L_LOW); // Опускаем (поднимаем) потенциал анода (катода) "отработавшего" сегмента
  if(++seg < 7) mask <<= 1; // Переходим к следующему сегменту
  else { // Дошли до последнего сегмента
    seg = 0; mask = 0x01;
  }
  digitalWrite(segPins[seg], L_HIGH); // Поднимаем (роняем) потенциал анода (катода) нового текущего сегмента
  for(int bi = 0; bi < 3; bi++) { // Теперь пробегаем по цифрам индикатора
    if(digit[d[bi]] & mask) // Если данный сегмент в bi-ом разряде должен гореть,
      digitalWrite(digPins[bi], L_LOW); // то опустим (поднимем) потенциал катода (анода) соответствующей цифры
    else // А если гореть не должен, то
      digitalWrite(digPins[bi], L_HIGH); // потенциал общего катода (анода) bi-й цифры должен быть высокий (низкий)
  }
}
//---------------------------------------------------------------------------------------
void receiveData(int byteCount) { // callback для принимаемых по i2c байтов
  while(Wire.available()) {
    stp = Wire.read(); stpChangedFlag = 1;
  }
}
//---------------------------------------------------------------------------------------
void sendData() { // callback для передаваемых по i2c байтов
    Wire.write(stp);
}
//---------------------------------------------------------------------------------------

